#!/bin/bash
# A program which manages Ubuntu-it's web test server
# Copyright (C) 2015  Pietro Albini <pietroalbini@ubuntu.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; witout even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib
import os
import sys
import json
import subprocess
import random
import string

import click

from . import app
from . import utils


DEFAULT_GIT_URL = "http://code.ubuntu-it.org/ubuntu-it-web/www.git"


def error(message, *format):
    """Show an error message"""
    click.echo("managetests: error: {}".format(message.format(*format)))
    sys.exit(1)


def random_string(length):
    """Generate a random string"""
    rand = random.SystemRandom()  # Uses /dev/urandom
    chars = string.ascii_letters+string.digits
    return "".join(rand.choice(chars) for i in range(length))


@click.group()
@click.option("-r", "--root", help="The managetests root")
@click.pass_context
def cli(ctx, root):
    """Manage a managetests instance"""
    ctx.obj = {}
    skip_check = ["init"]

    if root is not None and "MANAGETESTS_ROOT" in os.environ:
        root = os.environ["MANAGETESTS_ROOT"]

    skip = ctx.invoked_subcommand in skip_check
    if root is None or not (skip or utils.is_root_valid(root)):
        error("please provide a valid root path")
    else:
        ctx.obj["root"] = pathlib.Path(root)


@cli.command("init")
@click.argument("git-url", default=DEFAULT_GIT_URL)
@click.pass_context
def init_command(ctx, git_url):
    """Init a new root directory"""
    root = ctx.obj["root"]

    # Check if the directory is empty
    if root.exists() and len(list(root.iterdir())):
        error("root directory not empty")

    root.mkdir(parents=True)
    for dir in ("git", "branches"):
        (root / dir).mkdir()

    with (root / "version").open("w") as f:
        f.write("%s\n" % utils.ROOT_DIR_VERSION)

    with (root / "config.json").open("w") as f:
        content = {
            "gitlab-url": "http://code.ubuntu-it.org",
            "gitlab-project": "ubuntu-it-web/www",
            "gitlab-token": "abcdefghi",
            "hooks-token": random_string(32),
            "keep-branches": [
                "master",
            ],
        }
        json.dump(content, f, indent=4)
        f.write("\n")

    with (root / "details.json").open("w") as f:
        content = {
            "branches": {},
        }
        json.dump(content, f, indent=4)
        f.write("\n")

    subprocess.call(["git", "clone", git_url, str(root / "git"), "--bare"])


@cli.command("run")
@click.option("-p", "--port", help="The port for the frontend application",
            default=8080)
@click.pass_obj
def run_command(obj, port):
    """Run managetests"""
    root = obj["root"].resolve()

    with (root / "config.json").open() as f:
        config = json.load(f)

    inst = app.TestsManager(root, config, port)
    inst.run()
