# A program which manages Ubuntu-it's web test server
# Copyright (C) 2015  Pietro Albini <pietroalbini@ubuntu.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; witout even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import sys
import subprocess
import os
import shutil


BRANCH_DIR_VERSION = "1"


class Branch:
    """Representation of a branch"""

    def __init__(self, manager, name, merge_request):
        self.manager = manager

        self.name = name
        self.mr = merge_request
        self.mr_id = None
        self.mr_url = None
        self.author = None
        self.author_url = None
        self.assignee = None
        self.assignee_url = None

        self.pinned = name in manager.config["keep-branches"]

        self.check_remote_status()
        self.check_local_status()

    def check_remote_status(self):
        """Check the branch merging status on GitLab"""
        if self.pinned:
            self.active = True
            return

        result = self.manager.gitlab.merge_request(self.mr)
        if result:
            active = result["state"] == "opened"
        else:
            active = False
        self.active = active
        self.mr_id = result["iid"]
        self.mr_url = (self.manager.config["gitlab-url"]+"/"+
                       self.manager.config["gitlab-project"]+
                       "/merge_requests/"+str(self.mr_id))

        self.author = result["author"]["username"]
        self.author_url = result["author"]["web_url"]

        if result["assignee"] is not None:
            self.assignee = result["assignee"]["username"]
            self.assignee_url = result["assignee"]["web_url"]

    def check_local_status(self):
        """Check if the branch is present locally"""
        if (self.manager.root / "branches" / self.name).exists():
            self.present = self.valid()
        else:
            self.present = False

    def valid(self):
        """Check if a branch is valid"""
        root = self.manager.root / "branches" / self.name

        # The branch must be a valid directory
        if not (root.exists() and root.is_dir()):
            return False

        # All the subdirectories must be present
        dirs = ["env", "public", "data"]
        for dir in dirs:
            dir = root / dir
            if not (dir.exists() and dir.is_dir()):
                return False

        # All the files must be be present
        files = ["version"]
        for file in files:
            file = root / file
            if not (file.exists() and file.is_file()):
                return False

        # The directory version must be correct
        with (root / "version").open() as f:
            if f.read().strip() != BRANCH_DIR_VERSION:
                return False

        return True

    def deploy(self):
        """Deploy the branch"""
        self.check_local_status()

        # Ok, branch already deployed
        if self.present:
            return

        # Start from a fresh branch dir
        branch = self.manager.root / "branches" / self.name
        if branch.exists():
            shutil.rmtree(str(branch))
        branch.mkdir()

        git_dir = str(self.manager.root / "git")
        build_dir = branch / "build"

        # Create subdirectories
        subdirs = ["env", "public", "data", "build"]
        for dir in subdirs:
            (branch / dir).mkdir()

        # Create the version file
        with (branch / "version").open("w") as f:
            f.write("%s\n" % BRANCH_DIR_VERSION)

        # Update the bare git repository and checkout the branch
        subprocess.call(["git", "--git-dir", git_dir, "fetch", "origin"])
        subprocess.call([
            "git", "--git-dir", git_dir, "--work-tree", str(build_dir),
            "checkout", "-f", self.name,
        ])

        # Build the website
        env = dict(os.environ)
        env["HOME"] = str(build_dir)
        subprocess.call(["invoke", "build"], cwd=str(build_dir), env=env)

        # Check if the build results exists
        results = list(build_dir.glob("build/packages/*.whl"))
        if not results:
            return False

        # Create a new environment and install all the results in it
        subprocess.call(["virtualenv", "-p", "python3", str(branch / "env")])
        for result in results:
            subprocess.call([
                "%s/bin/pip" % str(branch / "env"), "install", str(result)
            ])

        # Calculate the static files directory
        py_version = ".".join(str(i) for i in tuple(sys.version_info)[:2])
        static_dir = (
            branch / "env" / "lib" / ("python%s" % py_version) /
            "site-packages" / "uitwww" / "static"
        )

        # Link the assets directory to the static files one
        assets = branch / "public" / "+assets"
        assets.symlink_to(static_dir)

        # Delete the build directory
        shutil.rmtree(str(build_dir))

        self.manager.details["branches"][self.name] = self.mr
        self.manager.save_details()

    def destroy(self):
        """Destroy the local copy"""
        self.check_local_status()

        # Ok, branch not present
        if not self.present:
            return

        shutil.rmtree(str(self.manager.root_dir / "branches" / name))

        try:
            del self.manager.details["branches"][self.name]
        except KeyError:
            pass
        self.manager.save_details()
