# A program which manages Ubuntu-it's web test server
# Copyright (C) 2015  Pietro Albini <pietroalbini@ubuntu.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; witout even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import queue
import threading
import logging

import flask


class HooksProcessor(threading.Thread):
    """A class which processes hooks"""

    def __init__(self, manager):
        self.manager = manager
        self.q = queue.Queue()
        self.stop = False

        super(HooksProcessor, self).__init__()

    def append(self, data):
        """Append something to the queue"""
        self.q.put(data)

    def run(self):
        """Run the thread"""
        while not self.stop:
            try:
                data = self.q.get(timeout=0.2)
            except queue.Empty:
                continue

            self.manager.process_hook(data)


def create_app(manager, processor):
    """Create an instance of the frontend app"""
    app = flask.Flask(__name__, static_url_path="/+assets")

    @app.errorhandler(404)
    def on_404(*_):
        return flask.render_template("404.html"), 404

    @app.route("/")
    def list_branches():
        branches = [b for b in manager.branches.values()
                    if b.name in manager.instances._processes]
        branches.sort(key=lambda branch: branch.name)
        branches.sort(key=lambda branch: branch.pinned, reverse=True)

        return flask.render_template("index.html", branches=branches)

    @app.route("/+hook/<token>", methods=["POST"])
    def hook(token):
        if token != manager.config["hooks-token"]:
            return "UNAUTHORIZED", 401

        processor.append(flask.request.json)
        return "OK", 200

    # Because it's impossible to run this app in debug mode, this handler
    # prints exceptions to the standard output
    handler = logging.StreamHandler()
    handler.setLevel(logging.INFO)
    app.logger.addHandler(handler)

    return app
