# A program which manages Ubuntu-it's web test server
# Copyright (C) 2015  Pietro Albini <pietroalbini@ubuntu.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published
# by the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; witout even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import subprocess
import atexit
import time

from werkzeug import serving


class InstancesManager:
    """This class will manage all the running branches"""

    def __init__(self, manager, main, main_port, processor):
        self.manager = manager
        self.running = False

        self._stop = False
        self._processes = {}
        self._branches = {}

        self._main_app = main
        self._main_port = main_port
        self._hooks_processor = processor

        atexit.register(_stop(self))

    def load_branch(self, branch):
        """Load a branch into the instances manager"""
        self._branches[branch.name] = branch

        if self.running:
            self._run_branch(branch)

    def remove_branch(self, name):
        """Remove a branch from the instances manager"""
        if name not in self._branches:
            return

        del self._branches[name]
        if self.running and name in self._processes:
            self._processes[name].terminate()
            del self._processes[name]

    def run(self):
        """Run all the instances"""
        if self.running:
            raise RuntimeError("Already running!")
        self.running = True
        self._stop = False

        for branch in self._branches.values():
            self._run_branch(branch)

        # Starts also the receiver part
        self._hooks_processor.start()
        serving.run_simple("127.0.0.1", self._main_port, self._main_app)
        self._hooks_processor.stop = True

        # Stop all the running processes
        for process in self._processes.values():
            process.terminate()
        self._processes = {}

        print()

    def _run_branch(self, branch):
        """Run a single branch"""
        branch_dir = self.manager.root / "branches" / branch.name
        command_path = str(branch_dir / "env" / "bin" / "uitwww")

        command = [
            command_path, "run", "-g",
            "%s.py" % self.manager.gunicorn_config_file,
            str(branch_dir / "data")
        ]
        try:
            process = subprocess.Popen(command)
        except FileNotFoundError:
            print("managetests: error: can't start the %s branch: executable "
                  "not found" % branch.name)
            return

        self._processes[branch.name] = process

    def stop(self):
        """Stop the server"""
        if not self.running:
            return

        self._stop = True


def _stop(inst):
    """Function passed to atexit which stops the know world"""
    def __(*_):
        if inst.running:
            inst.stop()
    return __
